from __future__ import annotations

import logging
import time
import warnings
from pathlib import Path

import numpy as np
import sklearn.preprocessing
from scipy.sparse import csr_matrix, sparray

from pymarkovclustering.logger import init_logger, init_null_logger, logging_timeit
from pymarkovclustering.utils import edges_to_sparse_matrix, extract_clusters

logger = logging.getLogger(__name__)


@logging_timeit(msg="Finished MCL")
def mcl(
    matrix: np.ndarray | csr_matrix,
    /,
    *,
    inflation: float = 2.0,
    max_iter: int = 100,
    quiet: bool = True,
) -> csr_matrix:
    """
    Run Markov Clustering

    Parameters
    ----------
    matrix : np.ndarray | csr_matrix
        Adjacency matrix
    inflation : float, optional
        Inflation factor
    max_iter : int, optional
        Max number of iterations
    quiet : bool, optional
        If True, pring log on screen

    Returns
    -------
    mcl_matrix : csr_matrix
        MCL result matrix

    Notes
    -----
    This function is designed to produce MCL results similar to `$ mcl abc.tsv --abc -I 2.0` command.
    See <https://micans.org/mcl/man/mcl.html> in detail.

    References
    ----------
    MCL - a cluster algorithm for graphs (<https://micans.org/mcl/>)
    """  # noqa: E501
    init_null_logger() if quiet else init_logger()

    row_num, col_num = matrix.shape
    if row_num != col_num:
        raise ValueError(f"Input matrix must be square ({row_num=}, {col_num=}).")

    matrix = to_undirected_sparse_matrix(matrix)

    logger.info(f"Starting MCL ({inflation=}, {max_iter=})")
    logger.info(f"Load {row_num}x{col_num} matrix with {len(matrix.data)} entries")
    set_self_loops(matrix)
    matrix = normalize(matrix)
    matrix = prune(matrix)
    # logger.info(f"Initial Normalization\n{matrix.toarray()}")
    for i in range(1, max_iter + 1):
        start_time = time.time()
        prev_matrix = matrix.copy()
        matrix = expand(matrix)
        matrix = inflate(matrix, inflation)
        matrix = prune(matrix)
        elapsed_time = time.time() - start_time
        logger.info(f"Iteration {i:03d} done --- {elapsed_time:.2f}[s]")
        # logger.info(f"After Expansion & Inflation\n{matrix.toarray()}")
        if allclose(matrix, prev_matrix):
            break
        if i == max_iter:
            warn = warnings.warn if quiet else logger.warning
            warn(f"MCL reached {max_iter=} and not converged yet. Consider increasing max_iter.")  # fmt: skip  # noqa: E501
    logger.info(f"{len(extract_clusters(matrix))} clusters found")
    return matrix


def easymcl(
    edges: str | Path | list[tuple[str, str, float]],
    /,
    *,
    inflation: float = 2.0,
    max_iter: int = 100,
    quiet: bool = True,
) -> list[list[str]]:
    """
    Run Markov Clustering from edges file or list of tuples

    easymcl automates load edges as matrix, MCL, extract clusters

    Parameters
    ----------
    edges : str | Path | list[tuple[str, str, float]]
        Edges(source, target, weight) file or list of tuples
    inflation : float, optional
        Inflation factor
    max_iter : int, optional
        Max number of iterations
    quiet : bool, optional
        If True, pring log on screen

    Returns
    -------
    clusters : list[list[str]]
        List of clusters, each cluster is a list of node labels

    Notes
    -----
    This function is designed to produce MCL results similar to `$ mcl abc.tsv --abc -I 2.0` command.
    See <https://micans.org/mcl/man/mcl.html> in detail.

    References
    ----------
    MCL - a cluster algorithm for graphs (<https://micans.org/mcl/>)

    Examples
    --------
    >>> import pymarkovclustering as pymcl
    >>> # easymcl automates load edges as matrix, MCL, extract clusters
    >>> clusters = pymcl.easymcl("edges.tsv")
    >>> # easymcl is same as code below
    >>> matrix, labels = pymcl.edges_to_sparse_matrix("edges.tsv")
    >>> mcl_matrix = pymcl.mcl(matrix)
    >>> clusters = pymcl.extract_clusters(mcl_matrix, labels)
    """  # noqa: E501
    matrix, labels = edges_to_sparse_matrix(edges)
    mcl_matrix = mcl(matrix, inflation=inflation, max_iter=max_iter, quiet=quiet)
    return extract_clusters(mcl_matrix, labels)


def to_undirected_sparse_matrix(matrix: np.ndarray | csr_matrix) -> csr_matrix:
    """
    Convert to undirected sparse matrix

    `A'ij = A'ji = Max(Aij, Aji)` [A = input matrix, A' = result matrix]

    Parameters
    ----------
    matrix : np.ndarray | csr_matrix
        Target matrix

    Returns
    -------
    matrix : csr_matrix
        Undirected sparse matrix
    """
    if isinstance(matrix, (np.ndarray, sparray)):
        matrix = csr_matrix(matrix)
    return matrix.maximum(matrix.T)


def set_self_loops(matrix: csr_matrix) -> None:
    """
    Set self-loops in matrix to max value of each column.
    If max column value is 0, set it to 1.

    Parameters
    ----------
    matrix : csr_matrix
        Target matrix
    """
    diag = matrix.max(axis=0).toarray().flatten()
    diag[diag == 0] = 1
    matrix.setdiag(diag)


def normalize(matrix: csr_matrix) -> csr_matrix:
    """
    Normalize matrix columns (each column sums to 1)

    Parameters
    ----------
    matrix : csr_matrix
        Target matrix

    Returns
    -------
    norm_matrix : csr_matrix
        Normalized matrix
    """
    return sklearn.preprocessing.normalize(matrix, norm="l1", axis=0)


def expand(matrix: csr_matrix) -> csr_matrix:
    """
    Expand matrix

    Parameters
    ----------
    matrix : csr_matrix
        Target matrix

    Returns
    -------
    expand_matrix : csr_matrix
        Expanded matrix
    """
    return matrix @ matrix


def inflate(matrix: csr_matrix, inflation_factor: float) -> csr_matrix:
    """
    Inflate matrix

    Parameters
    ----------
    matrix : csr_matrix
        Target matrix
    inflation_factor : float
        Inflation factor

    Returns
    -------
    inflate_matrix : csr_matrix
        Inflated matrix
    """
    return normalize(matrix.power(inflation_factor))


def prune(matrix: csr_matrix, threshold: float = 1e-4) -> csr_matrix:
    """
    Prune matrix value below threshold to 0

    Parameters
    ----------
    matrix : csr_matrix
        Target matrix
    threshold : float, optional
        Prune threshold

    Returns
    -------
    prune_matrix : csr_matrix
        Pruned matrix
    """
    matrix.data[matrix.data < threshold] = 0
    matrix.eliminate_zeros()
    return matrix


def allclose(
    m1: csr_matrix,
    m2: csr_matrix,
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> bool:
    """Returns True if two arrays are element-wise equal within a tolerance

    Parameters
    ----------
    m1, m2 : csr_matrix
        Matrix to compare
    rtol : float, optional
        Relative tolerance
    atol : float, optional
        Absolute tolerance

    Returns
    -------
    result : bool
        Check result
    """
    if m1.data.shape == m2.data.shape:
        return np.allclose(m1.data, m2.data, rtol=rtol, atol=atol)
    else:
        return False



from pathlib import Path

import pytest


@pytest.fixture()
def simple_edges() -> list[tuple[str, str, float]]:
    """Simple edges list of tuples"""
    return [
        ("A", "B", 10),
        ("B", "A", 8.0),
        ("A", "C", 10),
        ("B", "C", 2),
        ("D", "E", 5),
        ("F", "G", 2),
        ("H", "I", 0),
    ]


@pytest.fixture()
def simple_edges_file() -> Path:
    """Simple edges file"""
    return Path(__file__).parent / "resources" / "simple_edges.tsv"



import numpy as _np

from pymarkovclustering import logger as _logger
from pymarkovclustering.main import easymcl, mcl
from pymarkovclustering.utils import (
    edges_to_sparse_matrix,
    extract_clusters,
    random_edges,
    write_clusters,
    write_edges,
)
from pymarkovclustering.viz import easymclviz, mclviz

__version__ = "0.1.0"
__all__ = [
    "mcl",
    "easymcl",
    "edges_to_sparse_matrix",
    "extract_clusters",
    "random_edges",
    "write_edges",
    "write_clusters",
    "mclviz",
    "easymclviz",
]

_logger.init_null_logger()

# Better matrix display in jupyter notebook for debugging
_np.set_printoptions(
    edgeitems=10,
    linewidth=500,
    formatter={"float": lambda v: "   --" if v == 0 else f"{v:.3f}"},
)



import pymarkovclustering as pymcl


def test_mclviz(tmp_path):
    """Test pymclviz"""
    edges = pymcl.random_edges(100, random_add_rate=0.1, random_remove_rate=0.1)
    matrix, labels = pymcl.edges_to_sparse_matrix(edges)
    mcl_matrix = pymcl.mcl(matrix)
    clusters = pymcl.extract_clusters(mcl_matrix, labels)
    fig = pymcl.mclviz(matrix, labels, clusters)

    clusters_viz_file = tmp_path / "clusters.png"
    fig.savefig(clusters_viz_file, dpi=100)
    assert clusters_viz_file.exists()


def test_easymclviz(tmp_path):
    """Test pymclviz"""
    edges = pymcl.random_edges(100, random_add_rate=0.1, random_remove_rate=0.1)
    fig = pymcl.easymclviz(edges, show_label=True)

    clusters_viz_file = tmp_path / "clusters.png"
    fig.savefig(clusters_viz_file, dpi=100)  # type: ignore
    assert clusters_viz_file.exists()



from pymarkovclustering.utils import random_edges, write_clusters, write_edges


def test_write_edges(tmp_path):
    """Test write edges"""
    edges = [("A", "B", 0.5), ("B", "C", 1.0), ("C", "A", 0.8)]
    outfile = tmp_path / "edges.tsv"
    write_edges(outfile, edges)

    expected_content = "A\tB\t0.5\nB\tC\t1.0\nC\tA\t0.8"
    assert outfile.read_text(encoding="utf-8") == expected_content


def test_write_clusters(tmp_path):
    """Test write clusters"""
    clusters = [["A", "B"], ["C"], ["D", "E", "F"]]
    outfile = tmp_path / "clusters.tsv"
    write_clusters(outfile, clusters)

    expected_content = "A\tB\nC\nD\tE\tF"
    assert outfile.read_text(encoding="utf-8") == expected_content


def test_random_edges():
    """Test random edges generation"""
    node_size = 1000
    edges = random_edges(node_size, max_cluster_size=20)

    nodes = set()
    for source, target, weight in edges:
        assert 0.0 <= weight <= 1.0
        nodes.add(source)
        nodes.add(target)

    assert len(nodes) == node_size



from __future__ import annotations

import argparse
import logging
from pathlib import Path

import pymarkovclustering as pymcl
from pymarkovclustering.scripts import exit_handler

logger = logging.getLogger(__name__)


def main():
    """Main function called from CLI"""
    args = get_args()
    run(**args.__dict__)


@exit_handler
def run(
    edges: str | Path,
    *,
    outfile: str | Path | None,
    inflation: float = 2.0,
    max_iter: int = 100,
    quiet: bool = False,
) -> None:
    """
    Parameters
    ----------
    edges : str | Path
        Edges(source, target, weight) file
    outfile : str | Path | None
        Output cluster file
    inflation : float, optional
        Inflation factor
    max_iter : int, optional
        Max number of iteration
    quiet : bool, optional
        If True, print log on screen
    """
    clusters = pymcl.easymcl(
        edges,
        inflation=inflation,
        max_iter=max_iter,
        quiet=quiet,
    )
    if outfile is None:
        for cluster in clusters:
            print("\t".join(cluster))
    else:
        pymcl.write_clusters(outfile, clusters)
        logger.info(f"Write MCL clusters result to '{outfile}'")


def get_args() -> argparse.Namespace:
    """Get arguments

    Returns
    -------
    args : argparse.Namespace
        Argument parameters
    """
    description = "Markov Clustering in Python"
    parser = argparse.ArgumentParser(
        description=description,
        usage="pymcl [options] edges.tsv -o clusters.tsv",
        add_help=False,
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "edges",
        type=Path,
        help="Input edges(source, target, weight) tab-delimited file",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        help="Output tab-delimited clusters file (default: stdout)",
        default=None,
        metavar="",
    )
    default_inflation = 2.0
    parser.add_argument(
        "-I",
        "--inflation",
        type=float,
        help=f"Inflation factor (default: {default_inflation})",
        default=default_inflation,
        metavar="",
    )
    default_max_iter = 100
    parser.add_argument(
        "--max_iter",
        type=int,
        help=f"Max number of iteration (default: {default_max_iter})",
        default=default_max_iter,
        metavar="",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        help="No print log on screen (default: OFF)",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--version",
        version=f"v{pymcl.__version__}",
        help="Print version information",
        action="version",
    )
    parser.add_argument(
        "-h",
        "--help",
        help="Show this help message and exit",
        action="help",
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import sys
import time
from functools import partial, wraps
from pathlib import Path
from typing import Callable


def init_null_logger():
    """Initialize package root logger with NullHandler

    Configuring package root null logger for a library
    https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
    """
    pkg_root_name = __name__.split(".")[0]
    logger = logging.getLogger(pkg_root_name)
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logger.addHandler(logging.NullHandler())


def init_logger(
    *,
    quiet: bool = False,
    verbose: bool = False,
    log_file: str | Path | None = None,
):
    """Initialize package root logger with StreamHandler(& FileHandler)

    Configuring package root default logger for a CLI tool

    Parameters
    ----------
    quiet : bool, optional
        If True, no print info log on screen
    verbose: bool, optional
        If True & quiet=False, print debug log on screen
    log_file : str | Path | None, optional
        Log file
    """
    pkg_root_name = __name__.split(".")[0]
    logger = logging.getLogger(pkg_root_name)

    # Remove existing handler to avoid duplicate logging
    for handler in logger.handlers:
        logger.removeHandler(handler)
        handler.close()

    logger.setLevel(logging.DEBUG)
    log_formatter = logging.Formatter(
        fmt="$asctime | $levelname | $message",
        datefmt="%Y-%m-%d %H:%M:%S",
        style="$",
    )
    # Add stream handler for terminal stderr
    stream_handler = logging.StreamHandler(sys.stderr)
    stream_handler.setFormatter(log_formatter)
    if quiet:
        log_level = logging.WARNING
    else:
        log_level = logging.DEBUG if verbose else logging.INFO
    stream_handler.setLevel(log_level)
    logger.addHandler(stream_handler)

    if log_file:
        # Add file handler for log file
        file_handler = logging.FileHandler(log_file, mode="w", encoding="utf-8")
        file_handler.setFormatter(log_formatter)
        log_level = logging.DEBUG if verbose else logging.INFO
        file_handler.setLevel(log_level)
        logger.addHandler(file_handler)


def logging_timeit(
    func: Callable | None = None,
    /,
    *,
    msg: str = "Done",
    show_func_name: bool = False,
    debug: bool = False,
):
    """Elapsed time logging decorator

    e.g. `Done (elapsed time: 82.3[s]) [module.function]`

    Parameters
    ----------
    func : Callable | None, optional
        Target function
    msg : str, optional
        Logging message
    show_func_name : bool, optional
        If True, show elapsed time message with `module.function` definition
    debug : bool, optional
        If True, use `logger.debug` (By default `logger.info`)
    """
    if func is None:
        return partial(
            logging_timeit,
            msg=msg,
            show_func_name=show_func_name,
            debug=debug,
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        elapsed_time = time.time() - start_time
        logger = logging.getLogger(__name__)
        log = f"{msg} (elapsed time: {elapsed_time:.2f}[s])"
        if show_func_name:
            log = f"{log} [{func.__module__}.{func.__name__}]"
        logger_func = logger.debug if debug else logger.info
        logger_func(log)
        return result

    return wrapper



from __future__ import annotations

import logging
import signal
import sys
from functools import wraps


def exit_handler(func):
    """Exit handling decorator on exception

    The main purpose is logging on keyboard interrupt exception
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger(__name__)
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            logger.exception("Keyboard Interrupt")
            sys.exit(signal.SIGINT)
        except Exception as e:
            logger.exception(e)
            sys.exit(1)

    return wrapper



import numpy as np
from scipy.sparse import csr_matrix

import pymarkovclustering as pymcl
from pymarkovclustering.main import (
    allclose,
    expand,
    inflate,
    normalize,
    prune,
    set_self_loops,
    to_undirected_sparse_matrix,
)


def test_mcl_simple_edges(simple_edges):
    """Test MCL with simple input"""
    matrix, _ = pymcl.edges_to_sparse_matrix(simple_edges)
    mcl_matrix = pymcl.mcl(matrix, inflation=1.5, max_iter=100)
    assert np.array_equal(
        mcl_matrix.toarray(),
        np.array(
            [
                [1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            ],
        ),
    )


def test_easymcl_simple_edges(simple_edges):
    """Test easymcl with simple edges input"""
    clusters = pymcl.easymcl(simple_edges, inflation=1.5, max_iter=100)
    assert clusters == [["A", "B", "C"], ["D", "E"], ["F", "G"], ["H"], ["I"]]


def test_easymcl_simple_edges_file(simple_edges_file):
    """Test easymcl with simple edges file input"""
    clusters = pymcl.easymcl(simple_edges_file, inflation=2.0, max_iter=50, quiet=False)
    assert clusters == [["A", "B", "C"], ["D", "E"], ["F", "G"], ["H"], ["I"]]


def test_set_self_loops():
    """Test setting self-loops in matrix"""
    matrix = csr_matrix(
        np.array(
            [
                [1, 2, 0, 3],
                [4, 5, 0, 6],
                [7, 8, 0, 9],
                [0, 0, 0, 0],
            ]
        )
    )
    set_self_loops(matrix)
    assert np.array_equal(
        matrix.toarray(),
        np.array(
            [
                [7, 2, 0, 3],
                [4, 8, 0, 6],
                [7, 8, 1, 9],
                [0, 0, 0, 9],
            ]
        ),
    )


def test_normalize():
    """Test matrix normalization"""
    matrix = csr_matrix(
        np.array(
            [
                [5, 1, 1],
                [0, 1, 1],
                [0, 0, 3],
            ],
        ),
    )
    norm_matrix = normalize(matrix)
    assert np.array_equal(
        norm_matrix.toarray(),
        np.array(
            [
                [1.0, 0.5, 0.2],
                [0.0, 0.5, 0.2],
                [0.0, 0.0, 0.6],
            ],
        ),
    )


def test_expand():
    """Test matrix expansion"""
    matrix = csr_matrix(
        np.array(
            [
                [1.0, 0.5, 0.0],
                [0.0, 0.5, 0.5],
                [0.0, 0.0, 0.5],
            ],
        ),
    )
    expand_matrix = expand(matrix)
    assert np.array_equal(
        expand_matrix.toarray(),
        np.array(
            [
                [1.0, 0.75, 0.25],
                [0.0, 0.25, 0.50],
                [0.0, 0.00, 0.25],
            ],
        ),
    )


def test_inflate():
    """Test matrix inflation"""
    matrix = csr_matrix(
        np.array(
            [
                [0.5, 0.5],
                [1.0, 1.0],
            ],
        ),
    )
    inflate_matrix = inflate(matrix, inflation_factor=2)
    assert np.array_equal(
        inflate_matrix.toarray(),
        np.array(
            [
                [0.2, 0.2],
                [0.8, 0.8],
            ],
        ),
    )


def test_to_undirected_sparse_matrix():
    """Test directed to undirected sparse matrix"""
    matrix = np.array(
        [
            [1, 0, 0, 2],
            [0, 0, 0, 5],
            [3, 0, 1, 0],
            [1, 2, 1, 0],
        ],
    )
    undirected_matrix = to_undirected_sparse_matrix(matrix)
    assert np.array_equal(
        undirected_matrix.toarray(),
        np.array(
            [
                # M'ij=M'ji=Max(Mij, Mji)
                [1, 0, 3, 2],
                [0, 0, 0, 5],
                [3, 0, 1, 1],
                [2, 5, 1, 0],
            ],
        ),
    )


def test_prune():
    """Test sparse matrix pruning"""
    matrix = csr_matrix(
        np.array(
            [
                [1, 0, 0.1],
                [0, 1, 0.001],
                [1, 0.00001, 1],
            ]
        )
    )
    prune_matrix = prune(matrix, threshold=0.001)
    assert np.array_equal(
        prune_matrix.toarray(),
        np.array(
            [
                [1, 0, 0.1],
                [0, 1, 0.001],  # not pruned: 0.001 threshold boundary
                [1, 0, 1],  # pruned: 0.00001 < 0.001 threshold
            ]
        ),
    )


def test_allclose():
    """Test allclose function"""
    assert allclose(
        csr_matrix(np.array([[1, 2], [3, 4]])),
        csr_matrix(np.array([[1, 2], [3, 4]])),
    )
    assert not allclose(
        csr_matrix(np.array([[1, 2], [3, 4]])),
        csr_matrix(np.array([[1, 2], [3, 5]])),
    )



import shlex
import subprocess as sp


def test_pymcl_cli(simple_edges_file, tmp_path):
    """Test pymcl cli"""
    clusters_file = tmp_path / "clusters.tsv"

    cmd = f"pymcl {simple_edges_file} -o {clusters_file} -I 1.5"
    result = sp.run(shlex.split(cmd))

    assert result.returncode == 0
    assert clusters_file.exists()
    assert clusters_file.read_text(encoding="utf-8") == "A\tB\tC\nD\tE\nF\tG\nH\nI"


def test_pymcl__main__(simple_edges_file, tmp_path):
    """Test pymcl __main__ call"""
    cmd = f"python -m pymarkovclustering {simple_edges_file} -I 1.5"
    result = sp.run(shlex.split(cmd), capture_output=True, text=True, encoding="utf-8")

    assert result.returncode == 0
    assert result.stdout.strip() == "A\tB\tC\nD\tE\nF\tG\nH\nI"  # Check stdout



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from scipy.sparse import csr_matrix

import pymarkovclustering as pymcl

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure


def mclviz(
    matrix: np.ndarray | csr_matrix,
    labels: list[str],
    clusters: list[list[str]],
    /,
    *,
    ax: Axes | None = None,
    node_size: int = 20,
    node_cmap: str = "gist_rainbow",
    node_alpha: float = 1.0,
    edge_width: float = 1.0,
    edge_color: str = "lightgray",
    show_label: bool = False,
    label_va: str = "bottom",
    font_size: int = 8,
) -> Figure:
    """Visualize Markov Clustering clusters using networkx

    Parameters
    ----------
    matrix : np.ndarray | csr_matrix
        Adjacency matrix used as MCL input
    labels : list[str]
        Matrix labels
    clusters : list[list[str]]
        MCL clusters
    ax : Axes | None, optional
        Matplotlib axes. If None, auto created.
    node_size : int, optional
        Node plot size
    node_cmap : str, optional
        Node colormap (e.g. `gist_rainbow`, `jet`, `viridis`)
    node_alpha : float, optional
        Node color alpha parameter
    edge_width : float, optional
        Edge line width
    edge_color : str, optional
        Edge color
    show_label : bool, optional
        If True, show node label
    label_va : str, optional
        Node label vertical alignment (`top`|`center`|`bottom`|`baseline`|`center_baseline`)
    font_size : int, optional
        Node label size

    Returns
    -------
    fig : Figure
        Matplotlib figure

    Notes
    -----
    Additional installation of `networkx` and `matplotlib` are required for MCL clusters visualization.
    For better position layout, extra packages `pygraphviz`, `pydot`, `lxml` installation is preferred.
    See <https://networkx.org/documentation/stable/install.html> in details.
    """  # noqa: E501
    try:
        # Check extra module import
        import matplotlib.pyplot as plt
        import networkx as nx
    except ImportError:
        raise

    # Create matplotlib axes if ax is None
    if ax is None:
        fig = plt.figure(figsize=(8, 6), tight_layout=True)
        ax = fig.add_subplot()
        ax.axis(False)

    # Plot network graph
    G = nx.Graph(matrix)
    G.remove_edges_from(nx.selfloop_edges(G))
    try:
        # Better position layout using pygraphviz
        pos = nx.drawing.nx_agraph.pygraphviz_layout(G)
    except ImportError:
        pos = nx.drawing.spring_layout(G)
    nx.draw_networkx_nodes(
        G,
        pos,
        ax=ax,
        node_size=node_size,
        node_color=_get_node_colors(clusters, labels, node_cmap),  # type: ignore
        alpha=node_alpha,
    )
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=[(s, t) for s, t, d in G.edges.data() if d["weight"] > 0],
        ax=ax,
        width=edge_width,
        edge_color=edge_color,
    )
    if show_label:
        nx.draw_networkx_labels(
            G,
            pos,
            ax=ax,
            labels={i: label for i, label in enumerate(labels)},
            font_size=font_size,
            verticalalignment=label_va,
        )
    return ax.figure  # type: ignore


def easymclviz(
    edges: str | Path | list[tuple[str, str, float]],
    /,
    *,
    inflation: float = 2.0,
    max_iter: int = 100,
    quiet: bool = True,
    ax: Axes | None = None,
    node_size: int = 20,
    node_cmap: str = "gist_rainbow",
    node_alpha: float = 1.0,
    edge_width: float = 1.0,
    edge_color: str = "lightgray",
    show_label: bool = False,
    label_va: str = "bottom",
    font_size: int = 8,
) -> Figure:
    """Run Markov Clustering and visualize clusters using networkx

    easymclviz automates load edges as matrix, MCL, extract clusters, visualization

    Parameters
    ----------
    edges : str | Path | list[tuple[str, str, float]]
        Edges(source, target, weight) file or list of tuples
    inflation : float, optional
        Inflation factor
    max_iter : int, optional
        Max number of iterations
    quiet : bool, optional
        If True, pring log on screen
    ax : Axes | None, optional
        Matplotlib axes. If None, auto created.
    node_size : int, optional
        Node plot size
    node_cmap : str, optional
        Node colormap (e.g. `gist_rainbow`, `jet`, `viridis`, `tab20`)
    node_alpha : float, optional
        Node color alpha parameter
    edge_width : float, optional
        Edge line width
    edge_color : str, optional
        Edge color
    show_label : bool, optional
        If True, show node label
    label_va : str, optional
        Node label vertical alignment (`top`|`center`|`bottom`|`baseline`|`center_baseline`)
    font_size : int, optional
        Node label size

    Returns
    -------
    fig : Figure
        Matplotlib figure

    Notes
    -----
    Additional installation of `networkx` and `matplotlib` are required for MCL clusters visualization.
    For better position layout, extra packages `pygraphviz`, `pydot`, `lxml` installation is preferred.
    See <https://networkx.org/documentation/stable/install.html> in details.

    Examples
    --------
    >>> import pymarkovclustering as pymcl
    >>> # easymclviz automates load edges as matrix, MCL, extract clusters, visualization
    >>> fig = pymcl.easymclviz("edges.tsv")
    >>> # easymclviz is same as code below
    >>> matrix, labels = pymcl.edges_to_sparse_matrix("edges.tsv")
    >>> mcl_matrix = pymcl.mcl(matrix)
    >>> clusters = pymcl.extract_clusters(mcl_matrix, labels)
    >>> fig = pymcl.mclviz(matrix, labels, clusters)
    """  # noqa: E501
    matrix, labels = pymcl.edges_to_sparse_matrix(edges)
    mcl_matrix = pymcl.mcl(matrix, inflation=inflation, max_iter=max_iter, quiet=quiet)
    clusters = pymcl.extract_clusters(mcl_matrix, labels)
    return mclviz(
        matrix,
        labels,
        clusters,
        ax=ax,
        node_size=node_size,
        node_cmap=node_cmap,
        node_alpha=node_alpha,
        edge_width=edge_width,
        edge_color=edge_color,
        show_label=show_label,
        label_va=label_va,
        font_size=font_size,
    )


def _get_node_colors(
    clusters: list[list[str]],
    labels: list[str],
    cmap: str = "gist_rainbow",
) -> list[str]:
    """Get appropriate node colors from clusters & labels & cmap"""
    try:
        import matplotlib as mpl
        from matplotlib.colors import to_hex
    except:
        raise
    label2node_idx = {label: i for i, label in enumerate(labels)}
    node_idx2color = {}
    for cluster_idx, cluster in enumerate(clusters):
        for node_idx in map(label2node_idx.get, cluster):
            color = to_hex(mpl.colormaps[cmap](cluster_idx / (len(clusters) - 1)))
            node_idx2color[node_idx] = color
    return [v for _, v in sorted(node_idx2color.items())]



from __future__ import annotations

import csv
import itertools
import random
from collections import defaultdict
from itertools import combinations
from pathlib import Path

from scipy.sparse import csr_matrix


def edges_to_sparse_matrix(
    edges: str | Path | list[tuple[str, str, float]],
    /,
) -> tuple[csr_matrix, list[str]]:
    """Convert edges file or list of tuples to sparse matrix

    Parameters
    ----------
    edges : str | Path | list[tuple[str, str, float]]
        Edges(source, target, weight) file or list of tuples

    Returns
    -------
    matrix : csr_matrix
        Sparse matrix representation of edges
    nodes : list[str]
        List of node labels
    """
    # Load edges data
    edges = _load_edges(edges)

    # Prepare data for sparse matrix
    node2index = {}
    weights = []
    source_indices, target_indices = [], []
    for source, target, weight in edges:
        if source not in node2index:
            node2index[source] = len(node2index)
        if target not in node2index:
            node2index[target] = len(node2index)
        source_indices.append(node2index[source])
        target_indices.append(node2index[target])
        weights.append(weight)

    # Create sparse matrix
    nodes = list(node2index.keys())
    matrix = csr_matrix(
        (weights, (source_indices, target_indices)),
        shape=(len(nodes), len(nodes)),
    )
    return matrix, nodes


def _load_edges(
    edges: str | Path | list[tuple[str, str, float]],
) -> list[tuple[str, str, float]]:
    """Load edges file and return a list of tuples"""
    if isinstance(edges, (list, tuple)):
        return edges
    data: list[tuple[str, str, float]] = []
    with open(edges, encoding="utf-8") as f:
        reader = csv.reader(f, delimiter="\t")
        for row in reader:
            if len(row) >= 3:
                source, target, weight = row[0], row[1], row[2]
            else:
                source, target, weight = row[0], row[1], 1.0
            data.append((source, target, float(weight)))
    return data


def extract_clusters(
    matrix: csr_matrix,
    labels: list[str] | None = None,
    /,
) -> list[list[str]]:
    """
    Extract clusters from MCL result matrix and map them to labels

    Parameters
    ----------
    matrix : csr_matrix
        MCL result matrix
    labels : list[str] | None, optional
        List of labels corresponding to matrix indices.
        If None, `'0','1','2'...'X'` index label is used.

    Returns
    -------
    clusters : list[list[str]]
        List of clusters, where each cluster is a list of labels
    """
    if labels is None:
        labels = list(map(str, range(matrix.shape[0])))
    if len(labels) != matrix.shape[0]:
        raise ValueError("Length of labels must match matrix size")

    # Cluster labels based on row indices of max value in each column of matrix
    row_idx2cluster = defaultdict(list)
    for col_idx, row_idx in enumerate(matrix.argmax(axis=0).A1):  # type:ignore
        row_idx2cluster[row_idx].append(labels[col_idx])
    clusters: list[list[str]] = list(row_idx2cluster.values())

    # Check cluster labels match input labels
    cluster_labels = itertools.chain.from_iterable(clusters)
    if sorted(cluster_labels) != sorted(labels):
        raise ValueError("Input labels missing in result clusters. MCL bug? Failed to extract clusters.")  # fmt: skip  # noqa: E501

    return sorted(clusters, key=lambda c: len(c), reverse=True)


def random_edges(
    node_size: int = 100,
    /,
    *,
    min_cluster_size: int = 1,
    max_cluster_size: int = 10,
    min_weight: float = 0.0,
    max_weight: float = 1.0,
    random_add_rate: float = 0,
    random_remove_rate: float = 0,
    seed: int | None = 0,
) -> list[tuple[str, str, float]]:
    """
    Simple random edges generator for clustering test

    Node name is set as `{ClusterID}_{NodeID in Cluster}` e.g. `18_2`

    Parameters
    ----------
    node_size : int, optional
        Total number of nodes to generate
    min_cluster_size, max_cluster_size : int, optional
        Min-Max size of each cluster
    min_weight, max_weight : float, optional
        Min-Max weight for edges
    random_add_rate, random_remove_rate : float, optional
        Random add, remove edges rate for noisy dataset generation
    seed : int | None, optional
        Random seed for reproducibility

    Returns
    -------
    edges : list[tuple[str, str, float]]
        Edges list of tuples
    """  # noqa: E501
    random.seed(seed)
    if not 1 <= min_cluster_size <= max_cluster_size:
        raise ValueError(f"Invalid min-max cluster size (1 <= {min_cluster_size=} <= {max_cluster_size})")  # fmt: skip  # noqa: E501
    if not 0.0 <= min_weight <= max_weight:
        raise ValueError(f"Invalid min-max weight (0.0 <= {min_weight=} <= {max_weight=})")  # fmt: skip  # noqa: E501

    # Generate clusters until the total node size is reached
    data_count = 0
    cluster_count = 0
    clusters: list[list[str]] = []
    while True:
        cluster_count += 1
        cluster_size = random.randint(min_cluster_size, max_cluster_size)
        if data_count + cluster_size > node_size:
            cluster_size = node_size - data_count
        # Create a cluster with unique node labels
        cluster = []
        for i in range(1, cluster_size + 1):
            cluster.append(f"{cluster_count}_{i}")
        clusters.append(cluster)
        data_count += cluster_size
        # Stop if all nodes have been assigned to clusters
        if data_count == node_size:
            break

    # Generate edges list of tuples from clusters
    edges: list[tuple[str, str, float]] = []
    for cluster in clusters:
        if len(cluster) == 1:
            edges.append((cluster[0], cluster[0], 1.0))
        else:
            for s, t in combinations(cluster, 2):
                w = round(random.uniform(min_weight, max_weight), 3)
                edges.append((s, t, w))

    # Random remove edges
    edges = random.sample(edges, int(len(edges) * (1 - random_remove_rate)))
    # Random add edges
    labels = list(itertools.chain.from_iterable(clusters))
    for _ in range(int(len(edges) * random_add_rate)):
        s, t = random.sample(labels, 2)
        w = round(random.uniform(min_weight, max_weight) / 2, 3)
        edges.append((s, t, w))

    return edges


def write_edges(
    outfile: str | Path,
    edges: list[tuple[str, str, float]],
    /,
) -> None:
    """Write edges to file

    Parameters
    ----------
    outfile : str | Path
        Output tab-delimited edges file
    edges : list[tuple[str, str, float]]
        Edges list of tuples
    """
    with open(outfile, "w", encoding="utf-8") as f:
        f.write("\n".join(["\t".join((s, t, str(w))) for s, t, w in edges]))


def write_clusters(
    outfile: str | Path,
    clusters: list[list[str]],
    /,
) -> None:
    """Write clusters to file

    Parameters
    ----------
    outfile : str | Path
        Output tab-delimited clusters file
    clusters : list[list[str]]
        List of clusters, each cluster is a list of node labels
    """
    with open(outfile, "w", encoding="utf-8") as f:
        f.write("\n".join(["\t".join(c) for c in clusters]))
